/** $Header: cdi_lftmfrctn.c,v 1.13 99/05/26 16:21:49 jaaps Exp $
********************************** - - ****************************************
* 						   
*         _/_/_/_/_/_/    _/_/_/    _/_/      _/_/       SRON Utrecht
*      _/_/_/_/  _/_/  _/_/_/_/_/  _/_/_/  _/_/_/    
*     _/_/_/        _/_/      _/_/  _/_/_/_/_/       _/     _/ _/_/_/_/ _/_/_/
*      _/_/_/_/_/  _/_/_/_/_/_/_/    _/_/_/         _/     _/ _/     _/_/
*           _/_/_/_/_/_/_/_/_/_/  _/_/_/_/_/  _/_/ _/  _/ _/ _/_/_/ _/      
*  _/_/  _/_/_/_/_/_/      _/_/_/_/_/  _/_/_/     _/_/ _/_/ _/     _/_/  
* _/_/_/_/_/_/  _/_/      _/_/_/_/      _/_/     _/     _/ _/       _/_/_/   
*
*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
* Use	      :	cdi_lftmfrctn ();
*
* Parameters  :	
*
*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
* Dependencies:	
*
*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
* Description :	scan lifetime fraction data over the selected period,
*		compute a mean ltf.
*
*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
* $Log:	cdi_lftmfrctn.c,v $
 * Revision 1.13  99/05/26  16:21:49  16:21:49  jaaps (Jaap Schuurmans)
 * ** cdi_initltf: evt_sortlist, flag parameter was 0, need at least EVTFLG_DFLTTW to set
 * **              absolute inputfile time boundaries,
 * **		fixed.
 * 
 * Revision 1.12  99/05/21  12:28:01  12:28:01  jaaps (Jaap Schuurmans)
 *    cdi_initltf: do not try to open files named "-".
 * 
 * Revision 1.11  98/07/01  14:49:45  14:49:45  jaaps (Jaap Schuurmans)
 *    include <time.h>
 * 
* Revision 1.10  1998/06/29 14:03:27  jaaps
* no entry
*
* Revision 1.9  1998/06/25 13:41:59  jaaps
* non essential, cosmetic adjustments (no reason found in source).
*
 * Revision 1.8  1998/06/18  07:42:15  jaaps
 * non essential, cosmetic adjustments (no reason found in source).
 *
* Revision 1.7  1997/12/18 14:23:42  jaaps
*    fixed potentialproblem in ltf_copyname, null pntr deref.
*
* Revision 1.6  1997/09/22 11:47:23  jaaps
*    cdi_lftmfrctn: exp->ltffile key contents from EVL->LTF
*
* Revision 1.5  1997/09/17 13:58:43  jaaps
*    cdi_initltf: pdt files are allways level 1, transform prn0 --> pdt1
*
* Revision 1.4  1997/09/08 08:25:38  jaaps
*    the used livetime fraction filenames are now written to the FITS header.
*
* Revision 1.3  1997/07/04 11:34:40  jaaps
*    cdi_lftmfrctn: HEAD- added parameter for ltffile.
*    	          if names are provided, do no generate any inputfile names.
*
* Revision 1.2  1997/06/30 13:18:04  jaaps
*    cdi_lftmfrctn: bugfix, could return uninitialised lifetimefraction.
*
* Revision 1.1  1997/06/16 08:14:22  jaaps
* Initial revision
*
********************************** - - ****************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>

#include "wfcglobals.h"
#include "fits.h"
#include "util.h"

#include "instrument.h"

#include "expose.h"

#define C_CALLOC(s,a,b) secure_alloc (#s, a, b)

static char *RCSID = "$Header: cdi_lftmfrctn.c,v 1.13 99/05/26 16:21:49 jaaps Exp $";

static char *ltf_copyname (char *name) {
   
   char *rv, *s;
   
   rv= NULL;
   if (name != NULL && *name != '\0' && 
       strstr (name, "prn") != NULL  && 
       (rv= strdup (name))  != NULL) {
      if ((s = strstr (rv, "prn")) == NULL) 
	 wLog (DBG_FATAL, 
	       "ltf_copyname: this cannot be, see for yourself, .../src/cdi/"__FILE__"\n");
      strncpy (s, "pdt", 3); if (s[3] == '0') s[3]= '1';
   }
   return rv;
}

WFC_EVENTFILEPTR cdi_initltf (char *name, WFC_EVENTFILEPTR efp) {
   
   WFC_EVENTFILEPTR	ltf;
   char			mess[256];
   cardinal	 	i;
   
   ltf= evt_init (NULL, filter_cloneTime (efp->fltr), NULL, NULL, NULL, 0);
   if (ltf == NULL)
      wLog (DBG_FATAL, 
	    "cdi_initltf: INTERNAL ERROR, not supposed to happen, cf. cdi_initltf.c:""__LINE__""\n");
   
   if (name != NULL && *name != '\0') {
      ltf->nfiles = 0;
      if (strcmp (name, "-") != 0) {
	 ltf->source  = name;  
	 ltf->namelist= wft_resolvelist (ltf->source, &ltf->nfiles);
      }
   }
   else {
      ltf->nfiles= efp->nfiles;
      ltf->namelist= (char **) C_CALLOC(cdi_initltf, ltf->nfiles, sizeof (char *));   
      for (i= 0; i!= ltf->nfiles; i++)
	 ltf->namelist[i]= ltf_copyname (efp->namelist[i]);
   }
   evt_sortlist (ltf, EVTFLG_DFLTTW);
   if (ltf->nfiles == 0) {
      evt_clean (ltf);
      ltf= NULL;
   }
   
   if (ltf != NULL)
      if (ltf->mjde < efp->mjds ||
	  ltf->mjds > efp->mjde   ) {
	 wLog (DBG_ERROR, "cdi_initltf: event- and lifetimefraction data have no overlap.\n");
	 sprintf (mess,   "                event timespan: %.6f - %.6f\n", efp->mjds, efp->mjde);
	 wLog (DBG_CONT, mess);
	 sprintf (mess,   "     lifetimefraction timespan: %.6f - %.6f\n", ltf->mjds, ltf->mjde);
	 wLog (DBG_CONT, mess);
	 sprintf (mess,	  "             ignoring the ltf file%s, %s %s\n",
		  (ltf->nfiles > 1) ? "s" : "",
		  ltf->namelist[0],
		  (ltf->nfiles > 1) ? "....." : "");
	 wLog (DBG_CONT, mess);
	 evt_clean (ltf);
	 ltf= NULL;
      }	 
   
   if (ltf != NULL) {  
      if (ltf->wfc != 0 &&  ltf->wfc != efp->wfc) {
	 sprintf (mess,  
		  "cdi_initltf: \"%s\" lifetimefraction data for wfc%d, event data from wfc%d.\n",
		  ltf->evt->evtfile, ltf->wfc, efp->wfc);
	 wLog (DBG_ERROR, mess);
	 wLog (DBG_CONT, "********! MIXING DATA FROM BOTH Wide Field Camera's !********\n\n");
      }
      if (ltf->obsPeriod != NULL && efp->obsPeriod != NULL && 
	  strcmp (ltf->obsPeriod, efp->obsPeriod) != 0) {
	 sprintf (mess,  
		  "iros_effetime: lifetimefraction data from obs.period %s (\"%s\"),\n",
		  ltf->obsPeriod, ltf->namelist[0]);
	 wLog (DBG_ERROR, mess);
	 sprintf (mess,  
		  "               event data from obs.period %s (\"%s\").\n",
		  efp->obsPeriod, efp->namelist[0]);
	 wLog (DBG_CONT, mess);
	 wLog (DBG_CONT, "********! MIXING DATA FROM DIFFERENT OBSERVATION PERIODS !********\n\n");
      }
   }
   return ltf;
}
   

float cdi_lftmfrctn (WFC_EVENTFILEPTR *pltf, WFC_EXPOSEPTR exp, WFC_EVENTFILEPTR efp) {
   
   WFC_EVENTFILEPTR	ltf;
   RMD_LIFETIMEPTR	ltp;
   double		tt= 0.0, ett= 0.0, f;
   char			mess[256];
   cardinal		k, m;
   
   if (*pltf == NULL)
      *pltf= cdi_initltf (exp->ltffile, efp);
   else
      evt_reset (*pltf, NULL, NULL, NULL, 0, 0);

   ltf= *pltf;
   if (ltf != NULL) {
      exp->ltffile = "LTF";
      
      while ((ltp = (RMD_LIFETIMEPTR) ltf->eRead (NULL, ltf)) != NULL &&
	     ltp->mjd < efp->mjds)  ; /**skip till 1st tbin. **/
      
      do {
	 
	 ett+= ltp->ltf*ltp->igp; tt+= ltp->igp;
	 
      } while ((ltp = (RMD_LIFETIMEPTR) ltf->eRead (NULL, ltf)) != NULL &&
	       ltp->mjd < efp->mjde);
      
      exp->nltffile= 0;

      for (k= 0; k!= ltf->nfiles; k++) if (ltf->usedSet[k]) exp->nltffile++;
      
      if (exp->nltffile > 0) {
	 m= 0;
	 exp->ltflist= (char **) C_CALLOC(cdi_lftmfrctn, exp->nltffile, sizeof (char **));
	 for (k= 0; k!= ltf->nfiles; k++)
	    if (ltf->usedSet[k]) exp->ltflist[m++]= ltf->namelist[k];
      }
      
      f= 1.0;  if (tt > 0.0 && tt > ett) f= ett/tt;
      
      if ((gbl_debug&DBG_VISIBMASK) >= DBG_L2) {
	 time_t		st;
	 
	 st = (time_t) ((efp->mjds - asla_cldj ("1/1/70", "0:0:0"))*WFC_SECpDAY);
	 sprintf (mess, "cdi_lftmfrctn: start (UT) %s", asctime (gmtime (&st)) );
	 wLog (DBG_L2, mess);
	 sprintf (mess, "               accumulated effective exposure time %.3f sec.\n", ett);
	 wLog (DBG_CONT, mess);
	 sprintf (mess, "               accumulated   nominal exposure time %.3f sec.\n", tt);
	 wLog (DBG_CONT, mess);
	 sprintf (mess, "                             mean lifetimefraction %.3f\n"     , f);
	 wLog (DBG_CONT, mess);
      }	    
   }
   else {
      sprintf (mess, "cdi_lftmfrctn: could not establish a mean lifetimefraction.\n");
      wLog (DBG_WARNING, mess);
      f= 1.0;
   }
   return ((float) f);
}
