/** $Header: /home/sax/sources/DIST/src/evt/RCS/evt_read.c,v 1.13 1998/05/06 15:42:19 jaaps Exp $
********************************** - - ****************************************
* 						   
*         _/_/_/_/_/_/    _/_/_/    _/_/      _/_/       SRON Utrecht
*      _/_/_/_/  _/_/  _/_/_/_/_/  _/_/_/  _/_/_/    
*     _/_/_/        _/_/      _/_/  _/_/_/_/_/   _/     _/ _/_/_/_/ _/_/_/
*      _/_/_/_/_/  _/_/_/_/_/_/_/    _/_/_/     _/     _/ _/     _/_/
*           _/_/_/_/_/_/_/_/_/_/  _/_/_/_/_/   _/  _/ _/ _/_/_/ _/      
*  _/_/  _/_/_/_/_/_/      _/_/_/_/_/  _/_/_/ _/_/ _/_/ _/     _/_/  
* _/_/_/_/_/_/  _/_/      _/_/_/_/      _/_/ _/     _/ _/       _/_/_/   
*
*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
* Use	      :	
*
* Parameters  :	
*
*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
* Dependencies:	
*
*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
* Description :	
*
*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
* $Log: evt_read.c,v $
 * Revision 1.13  1998/05/06  15:42:19  jaaps
 *   evt_xrdatt: fixed bug in timewindow skip part,
 *   	      occasional out of array bounds access, causing coredumps.
 *
* Revision 1.12  1997/12/18 14:23:24  jaaps
* non essential, cosmetic adjustments (no reason found in source).
*
* Revision 1.11  1997/09/11 08:34:32  jaaps
* non essential, cosmetic adjustments (no reason found in source).
*
* Revision 1.10  1997/09/10 07:59:02  jaaps
* non essential, cosmetic adjustments (no reason found in source).
*
* Revision 1.9  1997/06/16 08:11:04  jaaps
*    extended the support for non-event type files.
*
* Revision 1.8  1997/06/04 15:21:41  jaaps
* non essential, cosmetic adjustments (no reason found in source).
*
* Revision 1.7  1997/04/11 15:21:08  jaaps
*   evt_xrdatt: cf. rev 1.5, this is OK, but do return correct flags in efp->state,
* 	      ie. state should not contain EVTFLG_ATTINW if there is no timewindow.
*
* Revision 1.6  1997/02/27 10:28:23  jaaps
*    evt_xrdatt: forward skip of attitude records in 'bad time interval'
*                could fail on buffer boundary, fixed.
* 	       short (less than attitude period) time intervals or short
* 	       gaps between intervals could be missed, fixed.
*
 * Revision 1.5  1997/01/30  13:45:37  jaaps
 *    evt_xrdatt:  if no explicit time window from user, reading attitude data was skipped.
 *    		though this initially looked oke, it may produce, SILENTLY, wrong results.
 * 		now allways honour the attitude data, if attitude file can be found.
 *
 * Revision 1.4  1996/12/18  11:24:47  jaaps
 *    added evt_xrdatt, to read attitude data.
 *
 * Revision 1.3  1996/12/13  12:35:31  jaaps
 * non essential, cosmetic adjustments (no reason found in source).
 *
 * Revision 1.2  1996/07/05  09:40:22  jaaps
 *    event read:  implemented funtions evt_xrdXXX, XXX= (nm, dm1, dm2)
 *    		attempt to be able to read any mode event lists,
 * 		specific read routine is in event.h:WFC_EVENTFILEREC.
 *      !!!! TAKE CARE - spv reader uses the older stuff.
 *
 * Revision 1.1  1996/04/25  21:03:46  sax
 * Initial revision
 *
********************************** - - ****************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "wfcglobals.h"
#include "dta.h"
#include "util.h"
#include "fits.h"

#include "event.h"

/**
*** #define C_CALLOC(a,b) calloc (a, b)
**/
#define C_CALLOC(a,b)  secure_alloc ("evt_read.c", a, b)


NM_EVENTPTR evt_evconvert (void *e, cardinal src) {
/** 
*** convert an event from type src (either DM1 or DM2) to NM.
***
***   
***if (src == WFC_EDM1ID) 
***   return evt_cvdr2nm (1, (DM1_EVENTPTR) e);
***else if (src == WFC_EDM2ID)
***   return NULL;
***else
**/
      return NULL;
}


void *evt_evread (WFC_EVENTFILEPTR efp, cardinal crc) {
   
   void *e= NULL;
   
   if (efp->n_tw == 0) efp->n_tw = evt_nexttimechunk (efp);
   
   if (efp->n_tw > 0) {
      
      switch (efp->f->ext->fileid) {
	 case  WFC_ENMID  : e= efp->anyp.pnme +efp->evcix; break;
	 case  WFC_EHTRID : e= efp->anyp.phtre+efp->evcix; break;
	 case  WFC_EBAMID : e= efp->anyp.pbame+efp->evcix; break;
	 case  WFC_EDM1ID : e= efp->anyp.pdm1e+efp->evcix; break;
	 case  WFC_EDM2ID : e= efp->anyp.pdm2e+efp->evcix; break;
	 case  WFC_EIHKID : e= efp->anyp.pihke+efp->evcix; break;
	 case  WFC_ESHKID : e= efp->anyp.pshke+efp->evcix; break;
	 default 	  : break;
      }
      efp->evcix+= 1;
      efp->n_tw -= 1;
      
      if (crc != 0 	&&
	  crc == WFC_ENMID &&
	  crc != efp->f->ext->fileid)
	 e= evt_evconvert (e, efp->f->ext->fileid);
   }
   return e;
}

NM_EVENTPTR evt_nmread (WFC_EVENTFILEPTR efp) {
      
   NM_EVENTPTR e;
   
   while ((e= (NM_EVENTPTR) evt_evread (efp, WFC_ENMID)) != NULL) {
      
      if (efp->fltr == NULL || filter_event (e, efp->fltr)) {
	 
	 if (efp->n_acc == 0) efp->t0 = e->time;
	 efp->t1 = e->time;
	 efp->n_acc++;
	 return e;
      }
      /** event did not pass filter, ignore. **/
   }
   return NULL;
}


DNM_EVENTPTR evt_dnmread (DNM_EVENTPTR e, WFC_EVENTFILEPTR efp) {
   
   float	x, y, t0;
   
   if (e == NULL)
      e= (DNM_EVENTPTR) C_CALLOC (1, sizeof (DNM_EVENTREC));
   
   while ((e->nmp= (NM_EVENTPTR) evt_evread (efp, WFC_ENMID)) != NULL) {
      
      e->fx = 0.0;
      e->fy = 0.0;
      
      if (efp->ebin != NULL) {
	 x  = ((short) e->nmp->x);
	 y  = ((short) e->nmp->y);
	 
	 t0= efp->ebin->r[0][0]*x + efp->ebin->r[0][1]*y + efp->ebin->o[0];
	 y = efp->ebin->r[1][0]*x + efp->ebin->r[1][1]*y + efp->ebin->o[1];
	 x = t0;
	 
	 if (x < 0.0) x= 0.0; if (y < 0.0) y= 0.0;
	 
	 e->fx = x-floor(x);
	 e->fy = y-floor(y);
	 
	 e->nmp->x= floor(x);
	 e->nmp->y= floor(y);
      }
      if (efp->fltr == NULL || filter_event (e->nmp, efp->fltr)) {
	 e->ix = e->nmp->x;
	 e->iy = e->nmp->y;
	 e->eix= (efp->fltr == NULL) ? 0 : efp->fltr->ch2bndmap[e->nmp->e];
	 
	 if (efp->n_acc == 0) efp->t0 = e->nmp->time;
	 efp->t1 = e->nmp->time;
	 efp->n_acc++;
	 
	 return e;
      }
   }
   return NULL;
}

void *evt_xrdnm  (void *e_in, WFC_EVENTFILEPTR efp) {
   
   DNM_EVENTPTR e= (DNM_EVENTPTR) e_in;
   NM_EVENTPTR	enm;
   float	x, y, t0;
   
   if (e == NULL) e= (DNM_EVENTPTR) C_CALLOC (1, sizeof (DNM_EVENTREC));
   
   loop :
      if (efp->n_tw == 0) efp->n_tw = evt_nexttimechunk (efp);
      
      if (efp->n_tw > 0) {
	 
	 e->nmp= efp->anyp.pnme +efp->evcix;
	 efp->evcix+= 1;
	 efp->n_tw -= 1;
	 
	 if (efp->fltr == NULL || filter_event (e->nmp, efp->fltr)) {
	    
	    e->fx = 0.0;  e->fy = 0.0;
	    
	    if (efp->ebin != NULL) {
	       x = ((short) e->nmp->x);    y = ((short) e->nmp->y);
	       
	       t0= efp->ebin->r[0][0]*x + efp->ebin->r[0][1]*y + efp->ebin->o[0];
	       y = efp->ebin->r[1][0]*x + efp->ebin->r[1][1]*y + efp->ebin->o[1];
	       x = t0;
	       
	       if (x < 0.0) x= 0.0; if (y < 0.0) y= 0.0;
	       
	       e->fx    = x-floor(x);  e->fy    = y-floor(y);	       
	       e->nmp->x= floor(x)  ;  e->nmp->y= floor(y);
	    }
	    e->ix = e->nmp->x;  e->iy = e->nmp->y;
	    e->eix= (efp->fltr 		  == NULL || 
		     efp->fltr->ch2bndmap == NULL    ) ? 0 : efp->fltr->ch2bndmap[e->nmp->e];
	    
	    if (efp->n_acc == 0) efp->t0 = e->nmp->time;
	    efp->t1 = e->nmp->time;
	    efp->n_acc++;
	    return (void *) e;
	 }
	 goto loop;
      }
      return NULL;
}

void *evt_xrddm1 (void *e_in, WFC_EVENTFILEPTR efp) {
   
   DNM_EVENTPTR e= (DNM_EVENTPTR) e_in;
   NM_EVENTPTR	enm;
   float	x, y, t0;
   bool		lme;
   
#ifdef SRON
   
   if (e == NULL)
      e= (DNM_EVENTPTR) C_CALLOC (1, sizeof (DNM_EVENTREC));
   
   loop :
      if (efp->n_tw == 0) efp->n_tw = evt_nexttimechunk (efp);
      
      if (efp->n_tw > 0) {
	 
	 lme= DSB_LEGALMAINEVENT ((efp->anyp.pdm1e+efp->evcix));
	 if (lme)
	    e->nmp= evt_dm1_to_nm (e->nmp, efp->anyp.pdm1e+efp->evcix);
	 else
	    efp->n_dmnlm++;
	 
	 efp->evcix+= 1;
	 efp->n_tw -= 1;
	 
	 if (!lme) goto loop;
	 
	 if (efp->fltr == NULL || filter_event (e->nmp, efp->fltr)) {
	    
	    e->fx = 0.0;
	    e->fy = 0.0;
	    
	    if (efp->ebin != NULL) {
	       x  = ((short) e->nmp->x);
	       y  = ((short) e->nmp->y);
	       
	       t0= efp->ebin->r[0][0]*x + efp->ebin->r[0][1]*y + efp->ebin->o[0];
	       y = efp->ebin->r[1][0]*x + efp->ebin->r[1][1]*y + efp->ebin->o[1];
	       x = t0;
	       
	       if (x < 0.0) x= 0.0; if (y < 0.0) y= 0.0;
	       
	       e->fx = x-floor(x);
	       e->fy = y-floor(y);
	       
	       e->nmp->x= floor(x);
	       e->nmp->y= floor(y);
	    }
	    e->ix = e->nmp->x;
	    e->iy = e->nmp->y;
	    e->eix= (efp->fltr 		  == NULL || 
		     efp->fltr->ch2bndmap == NULL    ) ? 0 : efp->fltr->ch2bndmap[e->nmp->e];
	    
	    if (efp->n_acc == 0) efp->t0 = e->nmp->time;
	    efp->t1 = e->nmp->time;
	    efp->n_acc++;
	    return (void *) e;
	 }
	 goto loop;
      }
#else
   wLog (DBG_ERROR, "evt_xrddm1: cannot handle dm1 eventfiles. (compile with -DSRON)\n");
      
#endif
   return NULL;
}

void *evt_xrddm2 (void *e_in, WFC_EVENTFILEPTR efp) {
   
   DNM_EVENTPTR e= (DNM_EVENTPTR) e_in;
   
   wLog (DBG_ERROR, "evt_xrddm2: cannot yet handle dm2 eventfiles.\n");
   return NULL;
}

void *evt_xrdatt (void *a_in /** a_in is dummy here, ignored **/, WFC_EVENTFILEPTR efp) {
   
   ATT_CAMERAPTR a;
   
   /**
   *** the attitude file spans the full observation.
   *** if we have no timewindow, scanning the attitude will always return all records
   *** in this period.
   *** usually, this is not wanted, the result is ,at least, puzzling and mostly
   *** wrong.
   *** However, this silent return on a NO TIMEWINDOW condition may leed to wrong
   *** results, as the open routines delete the tw chain if it concludes that indeed all
   *** records have to be read and it thus makes no sense to check timetags.
   **/
   
   if (efp->n_tw == 0) efp->n_tw = evt_nextattitude (efp);
   if (efp->n_tw == 0) return NULL;
   
   a= efp->anyp.patte+efp->evcix;
   
   efp->state &= ~EVTFLG_ATTBITS;			/** clear att-time status bits **/      
   
   if (efp->tw_fil != NULL) {
      efp->state |=  EVTFLG_ATTINW;			/** assume w're in a timewindow **/
      if (a->mjd < efp->tw_fil->r.s) { 			/** before tw.... skip attitude...  **/
	 double g, f;
	 cardinal n, nmx;
	 
	 f  = (double) efp->n_ext/efp->telaps; 	/** estimates #records per second **/
	 n  = (cardinal) (f*(efp->tw_fil->r.s - a->mjd)*WFC_SECpDAY);
	 nmx= efp->n_ext-efp->evcix;	 
	 if (n >= nmx) n= nmx-1;
	 
	 while (n >= efp->n_tw) {
	    n        -= efp->n_tw;  nmx      -= efp->n_tw;
	    if ((efp->n_tw = evt_nextattitude (efp)) == 0) return NULL;
	    efp->state |=  EVTFLG_ATTINW;
	    a           = efp->anyp.patte+efp->evcix;
	 }
	 
	 while (n < nmx-1 && a[n].mjd <  efp->tw_fil->r.s)
	    if (n < efp->n_tw)
	       n++;
	    else {			/** n>=efp->n_tw, crossing buffer boundary ... **/
	       n        -= efp->n_tw; nmx      -= efp->n_tw;
	       if ((efp->n_tw = evt_nextattitude (efp)) == 0) return NULL;
	       efp->state |=  EVTFLG_ATTINW;
	       a           = efp->anyp.patte+efp->evcix;
	    }	 
	 while (n > 0  && a[n].mjd > efp->tw_fil->r.s) n--;
	 
	 efp->evcix += n;  efp->n_tw  -= n;
	 
	 a= efp->anyp.patte+efp->evcix;	/** skip forward. **/
	 
	 efp->state |= EVTFLG_ATTTWEY; 				/** entering tw. **/
	 a->mjd= efp->tw_fil->r.s;
      }
      else  if (a->mjd >= efp->tw_fil->r.s+efp->tw_fil->r.L) {
	 efp->state |= EVTFLG_ATTTWEX;				/** exiting tw. **/
	 /***???? a->mjd= efp->tw_fil->r.s+efp->tw_fil->r.L;  ***/  
	 efp->tw_fil= efp->tw_fil->next;
	 if (efp->tw_fil == NULL) 				/** Last timewindow processed. **/
	    efp->state |= EVTFLG_EXPIRED;
	 if (efp->state&EVTFLG_EXPIRED)  return NULL;
      }
   }
   if (!(efp->state&EVTFLG_ATTTWEX)) {
      efp->evcix+= 1; efp->n_tw -= 1;
   }
   if (a != NULL 		&& 
       (gbl_debug&DBGMDL_EVT)	&&
       ((gbl_debug&DBG_VISIBMASK) >= DBG_L5)) {
      char mess[256];
      
      sprintf (mess, "evt_xrdatt: %.6f, geo[%.3f,%.3f,%.3f] angle %.3f\n",
	       a->mjd, a->geo_x, a->geo_y, a->geo_z, a->grimang);
      wLog (DBG_L4, mess);
   }

   return (void *) a;
}

void *evt_xrdltf  (void *l_in /** l_in is dummy here, ignored **/,
		   WFC_EVENTFILEPTR efp) 
{
   RMD_LIFETIMEPTR	l;
   
 loop:
    
   if (efp->n_tw == 0) efp->n_tw = evt_nextlifetime (efp);
   
   if (efp->n_tw > 0) {

      l= efp->anyp.pltfe +efp->evcix;
      
      
      if (efp->tw_fil != NULL) {
	 if (l->mjd <= efp->tw_fil->r.s) {
	    double   g, f;
	    cardinal n, nmx;
	    
	    f  = (double) efp->n_ext/efp->telaps;
	    n  = (cardinal) (f*(efp->tw_fil->r.s - l->mjd)*WFC_SECpDAY);
	    nmx= efp->n_ext-efp->evcix;	 
	    if (n >= nmx) n= nmx-1;
	    
	    if (n >= efp->n_tw) {
	       n        -= efp->n_tw;
	       if ((efp->n_tw = evt_nextlifetime (efp)) == 0) return (void *) NULL;
	       l           = efp->anyp.pltfe+efp->evcix;
	    }
	    
	    while (n < nmx && l[n].mjd <  efp->tw_fil->r.s) n++;
	    while (n > 0   && l[n].mjd >  efp->tw_fil->r.s) n--;
	    
	    efp->evcix += n;  l= efp->anyp.pltfe+efp->evcix;
	    efp->n_tw  -= n;
	 }
	 else  if (l->mjd > efp->tw_fil->r.s+efp->tw_fil->r.L) {
	    efp->tw_fil= efp->tw_fil->next;
	    if (efp->tw_fil == NULL) efp->n_tw= 0; /*<< fix, added 1-jun-99. */
	    goto loop;
	 }
      }
      
      efp->evcix+= 1;
      efp->n_tw -= 1;
      
      return (void *) l;
   }
   return (void *) NULL;
}
   
